"""
Toposcale functionalities

S. Filhol, Oct 2021

======= Organization of input data to Toposcale ======
dem_description (nb_points)
    X
    Y,
    elev,
    slope,
    aspect,
    svf,
    x_ERA,
    y_ERA,
    elev_ERA,

horizon_angles (nb_points * bins)
solar_geom (nb_points * time * 2)
surface_climate_data (spatial + time + var)
plevels_climate_data (spatial + time + var + plevel)


======= Dataset naming convention:  =========
ds_surf => dataset direct from ERA5 single level surface
ds_plev => dataset direct from ERA5 pressure levels

ds_surf_pt => 3*3 grid ERA5 single surface level around a given point (lat,lon)
ds_plev_pt => 3*3 grid ERA5 pressure level around a given point (lat,lon)

plev_interp => horizontal interpolation of the 3*3 grid at the point (lat,lon) for the pressure levels
surf_interp => horizontal interpolation of the 3*3 grid at the point (lat,lon) for the single surface level

top => closest pressure level above the point (lat,lon,elev) for each timesep
bot => closest pressure level below the point (lat,lon,elev) for each timesep

down_pt => downscaled data time series (t, u, v, q, LW, SW, tp)

TODO:
- check that transformation from r,t to q for plevels is working fine
- add metadata to all newly created variables in datasets ds_psurf, ds_plevel, down_pt
- upscale method for all points in df_centroids:
    - method (1) simply using a for loop over each row
    - method (2) concatenate 3*3 grid along an additional dimension to process all points in one go.

"""
import pdb

import pandas as pd
import xarray as xr
from pyproj import Transformer
import numpy as np
import sys, time
from TopoPyScale import meteo_util as mu
from multiprocessing.dummy import Pool as ThreadPool
import multiprocessing as mproc

# Physical constants
g = 9.81    #  Acceleration of gravity [ms^-1]
R = 287.05  #  Gas constant for dry air [JK^-1kg^-1]


def downscale_climate(path_forcing,
                      df_centroids,
                      horizon_da,
                      target_EPSG,
                      start_date,
                      end_date,
                      interp_method='idw',
                      lw_terrain_flag=True,
                      tstep='1H'):
    """
    Function to perform downscaling of climate variables (t,q,u,v,tp,SW,LW) based on Toposcale logic

    Args:
        path_forcing: path to forcing data [SURF*, PLEV*]
        df_centroids (dataframe): containing a list of point for which to downscale (includes all terrain data)
        horizon_da (dataarray): horizon angles for a list of azimuth
        target_EPSG (int): EPSG code of the DEM
        interp_method (str): interpolation method for horizontal interp. 'idw' or 'linear'
        lw_terrain_flag (bool): flag to compute contribution of surrounding terrain to LW or ignore
        tstep (str): timestep of the input data, default = 1H

    Returns:
        dataset: downscaled data organized with time, point_id, lat, long
    """
    print('\n---> Downscaling climate to list of points using TopoScale')
    start_time = time.time()
    tstep_dict = {'1H': 1, '3H': 3, '6H': 6}
    # =========== Open dataset with Dask =================
    tvec = pd.date_range(start_date, pd.to_datetime(end_date) + pd.to_timedelta('1D'), freq=tstep, closed='left')
    ds_plev = xr.open_mfdataset(path_forcing + 'PLEV*.nc', parallel=True).sel(time=tvec.values)
    ds_surf = xr.open_mfdataset(path_forcing + 'SURF*.nc', parallel=True).sel(time=tvec.values)

    # this block handles the expver dimension that is in downloaded ERA5 data if data is ERA5/ERA5T mix. If only ERA5 or
    # only ERA5T it is not present. ERA5T data can be present in the timewindow T-5days to T -3months, where T is today.
    # https://code.mpimet.mpg.de/boards/1/topics/8961
    # https://confluence.ecmwf.int/display/CUSF/ERA5+CDS+requests+which+return+a+mixture+of+ERA5+and+ERA5T+data
    try:
        # in case of there being an expver dimension and it has two or more values, select first value
        expverN = ds_plev["expver"].values[0]
        # create new datset when this dimoension only has a single value
        ds_plev = ds_plev.sel(expver=expverN)
        # finally this drops the coordinate
        ds_plev = ds_plev.drop("expver")
    except:
        print("No ERA5T  PRESSURE data present with additional dimension <expver>")

    try:
        # in case of there being an expver dimension and it has two or more values, select first value
        expverN = ds_surf["expver"].values[0]
        # create new datset when this dimoension only has a single value
        ds_surf = ds_surf.sel(expver=expverN)
        # finally this drops the coordinate
        ds_surf = ds_surf.drop("expver")
    except:
        print("No ERA5T SURFACE data present with additional dimension <expver>")

    # ============ Convert lat lon to projected coordinates ==================
    trans = Transformer.from_crs("epsg:4326", "epsg:" + str(target_EPSG), always_xy=True)
    nxv,  nyv = np.meshgrid(ds_surf.longitude.values, ds_surf.latitude.values)
    nlons, nlats = trans.transform(nxv, nyv)
    ds_surf = ds_surf.assign_coords({"latitude": nlats[:, 0], "longitude": nlons[0, :]})
    ds_plev = ds_plev.assign_coords({"latitude": nlats[:, 0], "longitude": nlons[0, :]})

    # ============ Loop over each point ======================================
    # Loop over each points (lat,lon) for which to downscale climate variable using Toposcale method

    solar_ds = xr.open_dataset('outputs/solar_ds.nc', chunks='auto', engine='h5netcdf')

    dataset = []
    dpt_list = []
    dpt_paths = []
    surf_paths  = []
    surf_list = []

    n_digits = len(str(df_centroids.index.max()))

    for i, row in df_centroids.iterrows():
        print('Downscaling t,q,u,v,tp,p for point: {} out of {}'.format(row.name+1, df_centroids.index.max()+1))
        # =========== Extract the 3*3 cells centered on a given point ============
        ind_lat = np.abs(ds_surf.latitude-row.y).argmin()
        ind_lon = np.abs(ds_surf.longitude-row.x).argmin()
        ds_surf_pt = ds_surf.isel(latitude=[ind_lat-1, ind_lat, ind_lat+1], longitude=[ind_lon-1, ind_lon, ind_lon+1])
        ds_plev_pt = ds_plev.isel(latitude=[ind_lat-1, ind_lat, ind_lat+1], longitude=[ind_lon-1, ind_lon, ind_lon+1])

        # ====== Horizontal interpolation ====================
        interp_method = 'idw'
        Xs, Ys = np.meshgrid(ds_plev_pt.longitude.values, ds_plev_pt.latitude.values)
        dist = np.sqrt((row.x - Xs)**2 + (row.y - Ys)**2)
        if interp_method == 'idw':
            idw = 1/(dist**2)
            weights = idw / np.sum(idw)  # normalize idw to sum(idw) = 1
        elif interp_method == 'linear':
            weights = dist / np.sum(dist)
        else:
            sys.exit('ERROR: interpolation method not available')

        # create a dataArray of weights to then propagate through the dataset
        da_idw = xr.DataArray(data = weights,
                              coords={
                                  "latitude": ds_plev_pt.latitude.values,
                                  "longitude": ds_plev_pt.longitude.values,
                              },
                              dims=["latitude", "longitude"]
                              )
        dw = xr.core.weighted.DatasetWeighted(ds_plev_pt, da_idw)
        plev_interp = dw.sum(['longitude', 'latitude'], keep_attrs=True)    # compute horizontal inverse weighted horizontal interpolation
        dw = xr.core.weighted.DatasetWeighted(ds_surf_pt, da_idw)
        surf_interp = dw.sum(['longitude', 'latitude'], keep_attrs=True)    # compute horizontal inverse weighted horizontal interpolation

        # ========= Converting z from [m**2 s**-2] to [m] asl =======
        plev_interp.z.values = plev_interp.z.values / g  # convert geopotential height to elevation (in m), normalizing by g
        plev_interp.z.attrs = {'units': 'm', 'standard_name': 'Elevation', 'Long_name': 'Elevation of plevel'}

        surf_interp.z.values = surf_interp.z.values / g  # convert geopotential height to elevation (in m), normalizing by g
        surf_interp.z.attrs = {'units': 'm', 'standard_name': 'Elevation', 'Long_name': 'Elevation of ERA5 surface'}

        # ============ Extract specific humidity (q) for both dataset ============
        surf_interp = mu.dewT_2_q_magnus(surf_interp, mu.var_era_surf)
        plev_interp = mu.t_rh_2_dewT(plev_interp, mu.var_era_plevel)

        if (row.elevation < plev_interp.z.isel(level=-1)).sum():
            print("---> WARNING: Point {} is {} m lower than the 1000hPa geopotential\n=> "
                  "Values sampled from Psurf and lowest Plevel. No vertical interpolation".
                  format(i, np.round(np.min(row.elevation - plev_interp.z.isel(level=-1).values),0)))
            ind_z_top = (plev_interp.where(plev_interp.z > row.elevation).z - row.elevation).argmin('level')
            top = plev_interp.isel(level=ind_z_top)
            down_pt = (top['t']).to_dataset()
            down_pt['u'] = top.u
            down_pt['v'] = top.v
            down_pt['q'] = top.q
            down_pt['tp'] = surf_interp.tp  * 1 / tstep_dict.get(tstep) * 10**3 # Convert to mm/hr
            down_pt['p'] = top.level*(10**2) * np.exp(-(row.elevation-top.z) / (0.5 * (top.t + down_pt.t) * R / g))  # Pressure in bar

        else:
            # ========== Vertical interpolation at the DEM surface z  ===============
            ind_z_bot = (plev_interp.where(plev_interp.z < row.elevation).z - row.elevation).argmax('level')
            ind_z_top = (plev_interp.where(plev_interp.z > row.elevation).z - row.elevation).argmin('level')
            top = plev_interp.isel(level=ind_z_top)
            bot = plev_interp.isel(level=ind_z_bot)

            # Preparing interpolation weights for linear interpolation =======================
            dist = np.array([np.abs(bot.z - row.elevation).values, np.abs(top.z - row.elevation).values])
            #idw = 1/dist**2
            #weights = idw / np.sum(idw, axis=0)
            weights = dist / np.sum(dist, axis=0)

            # ============ Creating a dataset containing the downscaled timeseries ========================
            down_pt = (bot.t * weights[1] + top.t * weights[0]).to_dataset()
            down_pt['u'] = bot.u * weights[1] + top.u * weights[0]
            down_pt['v'] = bot.v * weights[1] + top.v * weights[0]
            down_pt['q'] = bot.q * weights[1] + top.q * weights[0]
            down_pt['p'] = top.level*(10**2) * np.exp(-(row.elevation-top.z) / (0.5 * (top.t + down_pt.t) * R / g))  # Pressure in bar
            down_pt['tp'] = surf_interp.tp  * 1 / tstep_dict.get(tstep) * 10**3 # Convert to mm/hr

        # ======= logic  to compute ws, wd without loading data in memory, and maintaining the power of dask
        down_pt['theta'] = np.arctan2(-down_pt.u, -down_pt.v)
        down_pt['theta_neg'] = (down_pt.theta < 0)*(down_pt.theta + 2*np.pi)
        down_pt['theta_pos'] = (down_pt.theta >= 0)*down_pt.theta
        down_pt = down_pt.drop('theta')
        down_pt['wd'] = (down_pt.theta_pos + down_pt.theta_neg)  # direction in Rad
        down_pt['ws'] = np.sqrt(down_pt.u ** 2 + down_pt.v**2)
        down_pt = down_pt.drop(['theta_pos', 'theta_neg'])


        dpt_list.append(down_pt)
        dpt_paths.append('outputs/tmp/down_pt_{}.nc'.format(str(row.name).zfill(n_digits)))
        surf_list.append(surf_interp)
        surf_paths.append('outputs/tmp/surf_interp_{}.nc'.format(str(row.name).zfill(n_digits)))

        down_pt = None
        surf_interp = None

    xr.save_mfdataset(dpt_list, dpt_paths, engine='h5netcdf')
    dpt_list = None
    dpt_paths = None
    xr.save_mfdataset(surf_list, surf_paths, engine='h5netcdf')
    surf_list = None
    surf_paths = None

        
    ds_list = []
    ds_paths = []
    for i, row in df_centroids.iterrows():
        print('Downscaling LW, SW for point: {} out of {}'.format(row.name+1, df_centroids.index.max()+1))

        down_pt = xr.open_dataset('outputs/tmp/down_pt_{}.nc'.format(str(row.name).zfill(n_digits)), chunks='auto', engine='h5netcdf')
        surf_interp = xr.open_dataset('outputs/tmp/surf_interp_{}.nc'.format(str(row.name).zfill(n_digits)), chunks='auto', engine='h5netcdf')


        # ======== Longwave downward radiation ===============
        x1, x2 = 0.43, 5.7
        sbc = 5.67e-8

        down_pt = mu.mixing_ratio(down_pt, mu.var_era_plevel)
        down_pt = mu.vapor_pressure(down_pt, mu.var_era_plevel)
        surf_interp = mu.mixing_ratio(surf_interp, mu.var_era_surf)
        surf_interp = mu.vapor_pressure(surf_interp, mu.var_era_surf)

        # ========= Compute clear sky emissivity ===============
        down_pt['cse'] = 0.23 + x1 * (down_pt.vp / down_pt.t) ** (1 / x2)
        surf_interp['cse'] = 0.23 + x1 * (surf_interp.vp / surf_interp.t2m) ** (1 / x2)
        # Calculate the "cloud" emissivity, UNIT OF STRD (J/m2)
        surf_interp['cle'] = (surf_interp.strd/pd.Timedelta('1H').seconds) / (sbc * surf_interp.t2m**4) - surf_interp['cse']
        # Use the former cloud emissivity to compute the all sky emissivity at subgrid.
        surf_interp['aef'] = down_pt['cse'] + surf_interp['cle']
        if lw_terrain_flag:
            down_pt['LW'] = row.svf * surf_interp['aef'] * sbc * down_pt.t ** 4 + \
                            0.5 * (1 + np.cos(row.slope)) * (1 - row.svf) * 0.99 * 5.67e-8 * (273.15**4)
        else:
            down_pt['LW'] = row.svf * surf_interp['aef'] * sbc * down_pt.t ** 4

        kt = surf_interp.ssrd * 0
        sunset = solar_ds.sel(point_id=row.name).sunset
        mu0 = solar_ds.sel(point_id=row.name).mu0
        SWtoa = solar_ds.sel(point_id=row.name).SWtoa


        #pdb.set_trace()
        kt[~sunset] = (surf_interp.ssrd[~sunset]/pd.Timedelta('1H').seconds) / SWtoa[~sunset]     # clearness index
        kt[kt < 0] = 0
        kt[kt > 1] = 1
        kd = 0.952 - 1.041 * np.exp(-1 * np.exp(2.3 - 4.702 * kt))    # Diffuse index

        surf_interp['SW'] = surf_interp.ssrd/pd.Timedelta('1H').seconds
        surf_interp['SW'][surf_interp['SW'] < 0] = 0
        surf_interp['SW_diffuse'] = kd * surf_interp.SW
        down_pt['SW_diffuse'] = row.svf * surf_interp.SW_diffuse

        surf_interp['SW_direct'] = surf_interp.SW - surf_interp.SW_diffuse
        # scale direct solar radiation using Beer's law (see Aalstad 2019, Appendix A)
        ka = surf_interp.ssrd * 0
        #pdb.set_trace()
        ka[~sunset] = (g * mu0[~sunset]/down_pt.p)*np.log(SWtoa[~sunset]/surf_interp.SW_direct[~sunset])
        # Illumination angle
        down_pt['cos_illumination_tmp'] = mu0 * np.cos(row.slope) + np.sin(solar_ds.sel(point_id=row.name).zenith) *\
                                          np.sin(row.slope) * np.cos(solar_ds.sel(point_id=row.name).azimuth - row.aspect)
        down_pt['cos_illumination'] = down_pt.cos_illumination_tmp * (down_pt.cos_illumination_tmp > 0)  # remove selfdowing ccuring when |Solar.azi - aspect| > 90
        down_pt = down_pt.drop(['cos_illumination_tmp'])
        down_pt['cos_illumination'][down_pt['cos_illumination'] < 0 ] =0

        # Binary shadow masks.
        horizon = horizon_da.sel(x=row.x, y=row.y, azimuth=np.rad2deg(solar_ds.azimuth.isel(point_id=row.name)), method='nearest')
        shade = (horizon > solar_ds.sel(point_id=row.name).elevation)
        down_pt['SW_direct_tmp'] = down_pt.t * 0
        down_pt['SW_direct_tmp'][~sunset] = SWtoa[~sunset] * np.exp(-ka[~sunset] * down_pt.p[~sunset] / (g * mu0[~sunset]))
        down_pt['SW_direct'] = down_pt.t * 0
        down_pt['SW_direct'][~sunset] = down_pt.SW_direct_tmp[~sunset] * (down_pt.cos_illumination[~sunset] / mu0[~sunset]) * (1 - shade)
        down_pt['SW'] = down_pt.SW_diffuse + down_pt.SW_direct

        # currently drop azimuth and level as they are coords. Could be passed to variables instead.
        # round(5) required to sufficiently represent specific humidty, q (eg typical value 0.00078)
        down_pt = down_pt.drop(['level']).round(5)

        # adding metadata
        down_pt.LW.attrs = {'units': 'W/m**2', 'standard_name': 'Longwave radiations downward'}
        down_pt.cse.attrs = {'units': 'xxx', 'standard_name': 'Clear sky emissivity'}
        down_pt = down_pt.drop(['SW_direct_tmp'])
        down_pt.SW.attrs = {'units': 'W/m**2', 'standard_name': 'Shortwave radiations downward'}
        down_pt.SW_diffuse.attrs = {'units': 'W/m**2', 'standard_name': 'Shortwave diffuse radiations downward'}

        ds_list.append(down_pt)
        ds_paths.append('outputs/down_pt_{}.nc'.format(str(row.name).zfill(n_digits)))

        down_pt = None
        surf_interp = None
    xr.save_mfdataset(ds_list, ds_paths, engine='h5netcdf')
    ds_list = None
    ds_paths = None


        #down_pt.to_netcdf('outputs/down_pt_{}.nc'.format(row.name))
        #dataset.append(down_pt)

    # print timer to console
    print('---> Downscaling finished in {}s'.format(np.round(time.time()-start_time), 1))


def read_downscaled(path='outputs/down_pt*.nc'):
    """
    Function to read downscaled points saved into netcdf files into one single dataset using Dask

    Args:
        path (str): path and pattern to use into xr.open_mfdataset

    Returns:
        dataset: merged dataset readily to use and loaded in chuncks via Dask
    """
    down_pts = xr.open_mfdataset(path, concat_dim='point_id', combine='nested', parallel=True)
    return down_pts














"""
Collectio of plotting functions for TopoPyScale
S. Filhol, December 2021

"""

import matplotlib.pyplot as plt
from matplotlib.colors import LightSource

import numpy as np

def map__terrain(ds_param,
                     var='elevation',
                     hillshade=True,
                     ):
    '''
    Function to plot terrain parameters

    Args:
        ds_param:
        var:
        hillshade:

    Returns:

    '''
    plt.figure()
    alpha=1
    if hillshade:
        ls = LightSource(azdeg=315, altdeg=45)
        shade = ls.hillshade(ds_param.elevation.values, vert_exag=0.5,
                             dx=ds_param.x.diff('x')[0].values,
                             dy=ds_param.y.diff('y')[0].values,
                             fraction=1.0)


        plt.imshow(shade,
                   extent=[ds_param.x.min(), ds_param.x.max(), ds_param.y.min(), ds_param.y.max()],
                   cmap=plt.cm.gray)
        alpha=0.5
    plt.imshow(ds_param[var],
               cmap=cmap,
               extent=[ds_param.x.min(), ds_param.x.max(), ds_param.y.min(), ds_param.y.max()],
               alpha=alpha, **kwargs)
    plt.colorbar(label=var)
    plt.show()


def map_unclustered(ds_down,
                         ds_param,
                         time_step=1,
                         var='t',
                         cmap=plt.cm.RdBu_r,
                         hillshade=True,
                         **kwargs):
    """
    Function to plot unclustered downscaled points given that each point corresponds to a cluster label

    Args:
        ds_down:
        ds_param:
        time_step:
        var:
        cmap:
        **kwargs: kwargs to pass to imshow() in dict format.

    TODO:
    - if temperature, divergent colorscale around 0degC
    - see if we can remain within xarray dataset/dataarray to reshape and include x,y
    """

    ds_mini = ds_down.isel(time=time_step).copy()
    dc = ds_mini.sel(point_id=ds_param.cluster_labels.values.flatten())
    val = np.reshape(dc[var].values, ds_param.elevation.shape)

    plt.figure()
    alpha=1
    if hillshade:
        ls = LightSource(azdeg=315, altdeg=45)
        shade = ls.hillshade(ds_param.elevation.values, vert_exag=0.5,
                             dx=ds_param.x.diff('x')[0].values,
                             dy=ds_param.y.diff('y')[0].values,
                             fraction=1.0)


        plt.imshow(shade,
                   extent=[ds_param.x.min(), ds_param.x.max(), ds_param.y.min(), ds_param.y.max()],
                   cmap=plt.cm.gray)
        alpha=0.5
    plt.imshow(val,
               cmap=cmap,
               extent=[ds_param.x.min(), ds_param.x.max(), ds_param.y.min(), ds_param.y.max()],
               alpha=alpha, **kwargs)
    plt.colorbar()
    plt.show()



"""
Toposcale class definition

S. Filhol, September 2021

project/
    config.ini
    -> input/
        -> dem/
        -> climate/
    -> output/

"""
import os
#import configparser
import sys
import shutil
from munch import DefaultMunch
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from TopoPyScale import fetch_era5 as fe
from TopoPyScale import topo_param as tp
from TopoPyScale import topo_sub as ts
from TopoPyScale import fetch_dem as fd
from TopoPyScale import solar_geom as sg
from TopoPyScale import topo_scale as ta
from TopoPyScale import topo_export as te
from TopoPyScale import topo_plot as tpl
from TopoPyScale import topo_obs as tpo

class Topoclass(object):
    """
    A python class to bring the typical use-case of toposcale in a user friendly object
    """

    def __init__(self, config_file):
        
        try:
            with open(config_file, 'r') as f:
                self.config = DefaultMunch.fromYAML(f)
        except IOError:
                print('ERROR: config file does not exist. Check path.')

        # remove outputs directory because if results already exist this causes concat of netcdf files
        try:
            shutil.rmtree(self.config.project.directory + '/outputs/')
        except:
            os.makedirs('/'.join((self.config.project.directory, 'outputs/')))

        # remove output fsm directory
        try:
            shutil.rmtree(self.config.project.directory + '/fsm_sims/')
        except:
            print("no FSM directory to clean")

        # remove output fsm directory
        try:
            shutil.rmtree(self.config.project.directory + '/ensemble/')
        except:
            print("no ensemble directory to clean")
        # check if tree directory exists. If not create it
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/')))
        if not os.path.exists('/'.join((self.config.project.directory, 'outputs/'))):
            os.makedirs('/'.join((self.config.project.directory, 'outputs/')))
        
        self.config.climate.path = self.config.project.directory + 'inputs/climate/'
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/climate/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/climate')))
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/climate/tmp/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/climate/tmp')))
        if not os.path.exists('/'.join((self.config.project.directory, 'outputs/tmp/'))):
            os.makedirs('/'.join((self.config.project.directory, 'outputs/tmp')))

        self.config.dem.path = self.config.project.directory + '/inputs/dem/'
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/dem/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/dem')))

        self.config.dem.filepath = self.config.dem.path + self.config.dem.file
        if not os.path.isfile(self.config.dem.filepath):
            
            if self.config.project.extent is not None:
                self.config.project.extent = dict(zip(['latN', 'latS', 'lonW', 'lonE'], self.config.project.extent))
            else:
                print('ERROR: no extent provided. Must follow this format: [latN, latS, lonW, lonE]')
                sys.exit()

            fd.fetch_dem(self.config.dem.path, self.config.project.extent, self.config.dem.epsg, self.config.dem.file)            
        else:
            print('\n---> DEM file found')


        if self.config.project.extent is not None:
            extent_NSWE = self.config.project.extent
            self.config.project.extent = dict(zip(['latN', 'latS', 'lonW', 'lonE'], extent_NSWE))
        else:
            # little routine extracting lat/lon extent from DEM
            self.config.project.extent = tp.get_extent_latlon(self.config.dem.filepath, self.config.dem.epsg)

        print(self.config.project.extent)
        print('Project lat/lon extent:')
        print('\t----------------------------')
        print('\t|      North:{}          |\n\t|West:{}          East:{}|\n\t|      South:{}          |'.format(np.round(self.config.project.extent.get('latN'),1),
                                                        np.round(self.config.project.extent.get('lonW'),1),
                                                          np.round(self.config.project.extent.get('lonE'),1),
                                                          np.round(self.config.project.extent.get('latS'),1)))
        print('\t----------------------------')

        self.toposub = self.Toposub()
        self.plot = self.Plotting()
        self.solar_ds = None
        self.horizon_da = None

        # add here a little routinr doing chek on start and end date in config.yml compare to forcing in case

        self.toposub.dem_path = self.config.dem.filepath

        if self.config.project.climate.lower() == 'era5':
            self.get_era5()

    class Plotting:
        '''
        Sub-Class with plotting functions for topoclass object
        '''
        def __int__(self):
            self.ds_param = None
            self.ds_down = None

        def map_variable(self,
                         time_step=1,
                         var='t',
                         cmap=plt.cm.RdBu_r,
                         hillshade=True,
                         **kwargs):

            # add logic to check ds_down and ds_param exist.
            tpl.map_unclustered(self.ds_down,
                         self.ds_param,
                         time_step=time_step,
                         var=var,
                         cmap=cmap,
                         hillshade=hillshade,
                         **kwargs)
        def map_terrain(self, var='elevation', hillshade=True, **kwargs):
            tpl.map_terrain(self.ds_param,
                            var=var,
                            hillshade=hillshade,
                            **kwargs
                            )

        def map_center_clusters(self,
                                background_var='elevation',
                                cmap=plt.cm.viridis,
                                hillshade=True,
                                **kwargs):
            print('to be implemented')

        def cluster_map(self):
            print('To be implemented')

        def timeseries(self):
            print('To be implemented')

        def solar_geom(self):
            print('To be implemented')

        def horizon(self):
            print('To be implemented')



    class Toposub:
        """
        Sub-Class to initialize variables to store TopoSub variables
        """
        def __init__(self):
            self.dem_path = None
            self.ds_param = None
            self.df_centroids = None
            self.kmeans_obj = None
            self.scaler = None

        def plot_clusters_map(self, var='cluster_labels', cmap=plt.cm.hsv, figsize=(14, 10)):
            ts.plot_center_clusters(self.dem_path, self.ds_param, self.df_centroids, var=var, cmap=cmap, figsize=figsize)

        def write_landform(self):
            ts.write_landform(self.dem_path, self.ds_param)

    def compute_dem_param(self):
        self.toposub.ds_param = tp.compute_dem_param(self.config.dem.filepath)


    def extract_pts_param(self, method='nearest', **kwargs):
        """
        Function to use a list point as input rather than cluster centroids from DEM segmentation (topo_sub.py/self.clustering_dem()).

        Args:
            df (dataFrame):
            method (str): method of sampling
            **kwargs: pd.read_csv() parameters
        Returns:
        """
        self.toposub.df_centroids = pd.read_csv(self.config.project.directory + 'inputs/dem/' + self.config.sampling.points.csv_file, **kwargs)
        self.toposub.df_centroids = tp.extract_pts_param(self.toposub.df_centroids, self.toposub.ds_param, method=method)


    def extract_dem_cluster_param(self):
        """
        Function to segment a DEM in clusters and retain only the centroids of each cluster.
        :return:
        """
        df_param = ts.ds_to_indexed_dataframe(self.toposub.ds_param)
        df_scaled, self.toposub.scaler = ts.scale_df(df_param)
        if self.config.sampling.toposub.clustering_method.lower() == 'kmean':
            self.toposub.df_centroids, self.toposub.kmeans_obj, df_param['cluster_labels'] = ts.kmeans_clustering(
                df_scaled, 
                self.config.sampling.toposub.n_clusters, 
                seed=self.config.sampling.toposub.random_seed)
        elif self.config.sampling.toposub.clustering_method.lower() == 'minibatchkmean':
            self.toposub.df_centroids, self.toposub.kmeans_obj, df_param['cluster_labels'] = ts.minibatch_kmeans_clustering(
                df_scaled, 
                self.config.sampling.toposub.n_clusters, 
                self.config.project.CPU_cores,  
                seed=self.config.sampling.toposub.random_seed)
        else:
            print('ERROR: {} clustering method not available'.format(self.config.sampling.toposub.clustering_method))
        self.toposub.df_centroids = ts.inverse_scale_df(self.toposub.df_centroids, self.toposub.scaler)
        self.toposub.ds_param['cluster_labels'] = (["y", "x"], np.reshape(df_param.cluster_labels.values, self.toposub.ds_param.slope.shape))

        # update plotting class variable
        self.plot.ds_param = self.toposub.ds_param

    def extract_topo_param(self):
        """
        Function to select which 
        """
        if self.config.sampling.method == 'points':
            self.extract_pts_param()
        elif self.config.sampling.method == 'toposub':
            self.extract_dem_cluster_param()
        elif self.config.sampling.method == 'both':

            # implement the case one wann run both toposub and a list of points
            print('ERROR: method not yet implemented')
            
        else:
            print('ERROR: Extraction method not available')

    def compute_solar_geometry(self):
        self.solar_ds = sg.get_solar_geom(self.toposub.df_centroids,
                                          self.config.project.start,
                                          self.config.project.end,
                                          self.config.climate[self.config.project.climate].timestep,
                                          str(self.config.dem.epsg),
                                          self.config.project.CPU_cores)

    def compute_horizon(self):
        """
        Function to compute horizon angle and sample values for list of points
        :return:
        """
        self.horizon_da = tp.compute_horizon(self.config.dem.filepath, self.config.dem.horizon_increments, self.config.project.CPU_cores)
        tgt_x = tp.xr.DataArray(self.toposub.df_centroids.x.values, dims="points")
        tgt_y = tp.xr.DataArray(self.toposub.df_centroids.y.values, dims="points")
        for az in self.horizon_da.azimuth.values:
            self.toposub.df_centroids['hori_azi_'+str(az)] = self.horizon_da.sel(x=tgt_x,
                                                                            y=tgt_y,
                                                                            azimuth=az,
                                                                            method='nearest').values.flatten()

    def downscale_climate(self):
        ta.downscale_climate(self.config.climate.path,
                                        self.toposub.df_centroids,
                                        self.horizon_da,
                                        self.config.dem.epsg,
                                        self.config.project.start,
                                        self.config.project.end,
                                        self.config.toposcale.interpolation_method,
                                        self.config.toposcale.LW_terrain_contribution,
                                        self.config.climate[self.config.project.climate].timestep)
        self.downscaled_pts = ta.read_downscaled()

        # update plotting class variables
        self.plot.ds_down = self.downscaled_pts


            
    def get_era5(self):
        """
        Funtion to call fetching of ERA5 data
        TODO:
        - merge monthly data into one file (cdo?)- this creates massive slow down!
        """
        lonW = self.config.project.extent.get('lonW') - 0.4
        lonE = self.config.project.extent.get('lonE') + 0.4
        latN = self.config.project.extent.get('latN') + 0.4
        latS = self.config.project.extent.get('latS') - 0.4

        # retreive ERA5 surface data
        fe.retrieve_era5(
            self.config.climate[self.config.project.climate].product,
            self.config.project.start,
            self.config.project.end,
            self.config.climate.path,
            latN, latS, lonE, lonW,
            self.config.climate[self.config.project.climate].timestep,
            self.config.climate[self.config.project.climate].download_threads,
            surf_plev='surf'
            )
        # retrieve era5 plevels
        fe.retrieve_era5(
            self.config.climate[self.config.project.climate].product,
            self.config.project.start,
            self.config.project.end,
            self.config.climate.path,
            latN, latS, lonE, lonW, 
            self.config.climate[self.config.project.climate].timestep,
            self.config.climate[self.config.project.climate].download_threads,
            surf_plev='plev',
            plevels=self.config.climate[self.config.project.climate].plevels,
            )

    def get_WMO_observations(self):
        """
        Function to download and parse in-situ data from WMO database
        """
        lonW = self.config.project.extent.get('lonW') - 0.4
        lonE = self.config.project.extent.get('lonE') + 0.4
        latN = self.config.project.extent.get('latN') + 0.4
        latS = self.config.project.extent.get('latS') - 0.4

        df = pd.DataFrame()
        df['dates'] = pd.date_range(self.config.project.start,
                                    self.config.project.end, freq='M')
        df['month'] = df.dates.dt.month
        df['year'] = df.dates.dt.year

        bbox = [latS, lonW, latN, lonE]
        tpo.fetch_WMO_insitu_observations(list(df.year.unique()),
                                          list(df.month.unique()))
        tpo.parse_WMO_insitu_observations()



    def to_cryogrid(self, fname_format='Cryogrid_pt_*.nc', precip_partition='continuous'):
        """
        wrapper function to export toposcale output to cryosgrid format from TopoClass
        
        Args:
            fname_format (str): filename format. point_id is inserted where * is

        """
        path = self.config.project.directory + 'outputs/'
        if 'cluster:labels' in self.toposub.ds_param.keys():
            label_map = True
            da_label = self.toposub.ds_param.cluster_labels
        else:
            label_map = False
            da_label = None
        te.to_cryogrid(self.downscaled_pts,
                       self.toposub.df_centroids,
                       fname_format=fname_format,
                       path=path,
                       label_map=label_map,
                       snow_partition_method=precip_partition,
                       da_label=da_label,
                       climate_dataset_name=self.config.project.climate,
                       project_author=self.config.project.authors)
        
    def to_fsm(self, fname_format='./outputs/FSM_pt_*.txt'):
        """
        function to export toposcale output to FSM format
        """
        te.to_fsm(self.downscaled_pts, fname_format)


    def to_crocus(self, fname_format='./outputs/CROCUS_pt_*.nc', scale_precip=1):
        """
        function to export toposcale output to crocus format .nc. This functions saves one file per point_id
        
        Args:
            fout_format (str): filename format. point_id is inserted where * is
            scale_precip(float): scaling factor to apply on precipitation. Default is 1
        """
        te.to_crocus(self.downscaled_pts,
                     self.toposub.df_centroids,
                     fname_format=fname_format,
                     scale_precip=scale_precip,
                     climate_dataset_name=self.config.project.climate,
                     project_author=self.config.project.authors)

    
    def to_snowmodel(self, fname_format='./outputs/Snowmodel_stn_*.csv'):
        """
        function to export toposcale output to snowmodel format .ascii, for single station standard

            fout_format: str, filename format. point_id is inserted where * is
        """
        te.to_micromet_single_station(self.downscaled_pts, self.toposub.df_centroids, fname_format=fname_format, na_values=-9999, headers=False)
    
    def to_netcdf(self, file_out='./outputs/output.nc'):
        """
        function to export toposcale output to one single generic netcdf format, compressed

        Args:
            file_out (str): name of export file
        """
        encod_dict = {}
        for var in list(self.downscaled_pts.keys()):
            scale_factor, add_offset = te.compute_scaling_and_offset(self.downscaled_pts[var], n=10)
            encod_dict.update({var:{"zlib": True,
                                   "complevel": 9,
                                   'dtype':'int16',
                                   'scale_factor':scale_factor,
                                   'add_offset':add_offset}})
        self.downscaled_pts.to_netcdf(file_out, encoding=encod_dict)
        print('---> File {} saved'.format(file_out))

    def to_snowpack(self, fname_format='./outputs/smet_pt_*.smet'):
        """
        function to export toposcale output to FSM format
        """
        te.to_snowpack(self.downscaled_pts, fname_format)

    def to_geotop(self, fname_format='./outputs/meteo_*.txt'):
        """
        function to export toposcale output to FSM format
        """
        te.to_geotop(self.downscaled_pts, fname_format)


    def to_musa(self, 
                fname_met='musa_met.nc', 
                fname_labels='musa_labels.nc'):
        """
        function to export TopoPyScale output in a format compatible with MuSa
        MuSa: https://github.com/ealonsogzl/MuSA

        Args:
            fname: filename of the netcdf 
        """

        te.to_musa(ds=self.downscaled_pts,
            df_pts=self.toposub.df_centroids,
            da_label=self.toposub.ds_param.cluster_labels,
            fname_met=fname_met,
            fname_labels=fname_labels,
            path=self.config.project.directory + 'outputs/',
            climate_dataset_name=self.config.project.climate,
            project_authors=self.config.project.authors
            )
        print('---> File ready for MuSa {} saved'.format(fname_met))